#! /usr/bin/python
# -*- mode: c; c-basic-offset: 2; indent-tabs-mode: nil; -*-
# vim:expandtab:shiftwidth=2:tabstop=2:smarttab:
#
# Copyright (C) 2009 Sun Microsystems
#
# Authors:
#
#  Jay Pipes <joinfu@sun.com>
#  Lee Bieber <lee.bieber@sun.com>
#
# This file is part of the Drizzle Automation Project.
#
# The Drizzle Automation Project is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation.
#
# The Drizzle Automation Project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with The Drizzle Automation Project (see COPYING.LESSER). If not, 
# see <http://www.gnu.org/licenses/>.

"""Script to automate running and processing of GCOV/LCOV data."""

import sys
import os
import os.path
from drizzle.automation.lib  import logging
import commands
import datetime

# print out failing tests if we encounter test failures
def check_test_suite_logs():
  infilename= "tests/var/log/drizzle-test-run.log"
  inf= open(infilename, "r+b")

  inlines= inf.readlines()
  in_ignore_block= False
  for inline in inlines:
    if inline.startswith("[ fail ]"):
      in_ignore_block= True
    if inline.startswith("Stopping All Servers") and in_ignore_block == True:
      in_ignore_block= False
      continue
    if in_ignore_block == True:
      logging.error(inline.strip())

  inf.close()

# Strip out include files that we are not interested in
def strip_include_files(working_dir):
  logging.info("Stripping LCOV files of unwanted includes.")
  ignore_patterns= [
    'SF:/usr/include/c++'
  , 'SF:/usr/local/include/google/protobuf'
  , 'SF:/usr/include/google/protobuf'
  , 'SF:/usr/include/security'
  ]
  os.chdir("%s/lcov" % working_dir)
  outfilename= "all.out.new"
  infilename= "all.out"
  inf= open(infilename, "r+b")
  outf= open(outfilename, "w+b")

  outlines= []
  inlines= inf.readlines()
  in_ignore_block= False
  for inline in inlines:
    for ignore_pattern in ignore_patterns:
      if inline.startswith(ignore_pattern):
        in_ignore_block= True
        break
    if inline.startswith("end_of_record") and in_ignore_block == True:
      in_ignore_block= False
      continue
    if in_ignore_block == False:
      outf.write(inline)

  inf.close()
  outf.flush()
  outf.close()
  os.rename(outfilename, infilename)

# (ugly) hack to get output to show correctly
# walk the drizzle source directory and get a list of files in the .libs directory
# then move the .gcda and .gcno files up one level
def move_lcov_files(working_dir):
  import distutils.file_util 
  logging.info("Moving .gcda and .gcno files from .libs directory up one directory level")
  for root,dirs,files in os.walk(working_dir, topdown=True):
    for d in dirs:
      if d == '.libs':
        dir= os.path.join("%s/%s" % (root,d))
        for next_root,next_dirs,next_files in os.walk(dir, topdown=True):
          for f in next_files:
            file= os.path.join("%s/%s" % (dir,f))
            # now find .gcda or .gcno files 
            if file.endswith(".gcda") or file.endswith(".gcno"):
              upper_dir = os.path.dirname(dir)
              # then move them up one level in the directory tree
              #logging.info("Moving %s to %s" % (file,upper_dir))
              distutils.file_util.move_file(file,upper_dir)

# ugly hack #2
# need to munge the string for some files in the plugin and drizzled directory
# as gcov can't find the source file and thus assumes incorrectly
# that it is in the root directory. So we need to add the missing path
# to the directory string
def patch_drizzled(working_dir):
  # this is the missing string
  working_dir2= os.path.dirname(working_dir)
  # this is what they look like before
  bad_strings= [
   'SF:' + working_dir +  '/drizzled/replication.pb.h'
  ,'SF:' + working_dir2 + '/drizzled/message/replication.pb.h'
  ,'SF:' + working_dir +  '/drizzled/table.pb.h'
  ,'SF:' + working_dir2 + '/drizzled/message/table.pb.h'
  ]

  # this is what they need to look like afterwards
  patch_strings= [
   'SF:' + working_dir + '/drizzled/message/replication.pb.h'
  ,'SF:' + working_dir + '/drizzled/message/replication.pb.h'
  ,'SF:' + working_dir + '/drizzled/message/table.pb.h'
  ,'SF:' + working_dir + '/drizzled/message/table.pb.h'
  ]
  patch_file_strings(working_dir, bad_strings, patch_strings, './drizzled/drizzled.out')

def patch_plugin(working_dir):
  # this is the missing string
  add_dir_string= 'plugin/innobase/pars'
  # this is what they look like before
  bad_strings= [
   'SF:' + working_dir + '/lexyy.c'
  ,'SF:' + working_dir + '/pars0grm.c'
  ,'SF:' + working_dir + '/pars0grm.y'
  ,'SF:' + working_dir + '/pars0lex.l'
  ]

  # this is what they need to look like afterwards
  patch_strings= [
   'SF:' + working_dir + '/' + add_dir_string + '/lexyy.c'
  ,'SF:' + working_dir + '/' + add_dir_string + '/pars0grm.c'
  ,'SF:' + working_dir + '/' + add_dir_string + '/pars0grm.y'
  ,'SF:' + working_dir + '/' + add_dir_string + '/pars0lex.l'
  ]
  patch_file_strings(working_dir, bad_strings, patch_strings, './plugin/plugin.out')

def patch_file_strings(working_dir, bad_strings, patch_strings, infilename):
  os.chdir(working_dir)
  outfilename= "/tmp/lcov.tmp"
  inf= open(infilename, "r+b")
  outf= open(outfilename, "w+b")

  inlines= inf.readlines()
  for inline in inlines:
    if inline.rstrip() in bad_strings:
      for index, string in enumerate(bad_strings):
        if inline.rstrip() == string:
          outf.write(patch_strings[index]+'\n')
          break
    else:
      outf.write(inline)

  inf.close()
  outf.flush()
  outf.close()
  os.rename(outfilename, infilename)

def execute(processing_mode, variables):
  # Set/verify some required variables, depending on the server
  # we are processing

  working_dir= variables['working_dir']
  run_date= datetime.datetime.now().isoformat()

  server_name= variables['server']

  os.chdir(working_dir)

  # Set of directories we're interested in running coverage reports on
  lcov_dirs= variables['lcov']['directories'].split(',')

  # For BZR mode, the "version" of the server (used in identifying the 
  # command run when logging to the database) is the branch and revision number.
  #
  # For MySQL Sandbox mode, the version is the text after the "msb_" prefix
  # on the sandbox.
  if processing_mode == 'bzr':
    server_version= "%s-%s" % (variables['bzr_branch'], variables['bzr_revision'])
  else:
    server_version= variables['mysql_sandbox'].split('_')[1] # Sandbox names are msb_XXXX (e.g. msb_6011 for MySQL 6.0.11)

  # Reset all coverage counters
  logging.info("Resetting all LCOV coverage counters.")

  for dir in lcov_dirs:
    cmd= "lcov --directory %s/%s --zerocounters" % (working_dir, dir)
    (retcode, output)= commands.getstatusoutput(cmd)
    if not retcode == 0:
      logging.error("Failed to zero-out coverage counters for %s/%s. Exiting." % (working_dir, dir))
      sys.exit(1)

  # Depending on the type of server we are testing/processing, we may
  # instantiate a "Builder" object, which is used to build the 
  # server from source...
  if server_name == 'drizzled':

    if variables['no_build'] is False:
      from drizzle.automation.builder import drizzled as builder_adapter
      if "make_options" in variables['defaults'].keys():
        make_options= variables['defaults']['make_options']
      else:
        make_options= "-j2"
      configure_options= '--with-debug --enable-coverage --enable-profiling'
      builder= builder_adapter.DrizzledBzrBuilder(variables['working_dir'], configure_options, make_options)
      builder.build()

  elif server_name == 'mysqld':

    pass # Not currently building mysql from source...

  else:
    logging.error("Not yet implemented!")
    sys.exit(1)

  # Run the test suite
  logging.info("Running test suite.")

  (retcode, ignored)= commands.getstatusoutput("make test")

  if retcode != 0:
    logging.error("Test suite failed to pass all tests. Exiting.")
    check_test_suite_logs()
    sys.exit(1)

  # Produce coverage info files
  logging.info("Producing LCOV info files for all directories.")
  
  # move lcov file out of .libs directory before capturing lcov info
  move_lcov_files(working_dir)

  for dir in lcov_dirs:
    # special case drizzled's --base-directory 
    if dir == 'drizzled':
      base_dir= '.'
    else: 
      base_dir= '..'
    logging.info("Processing %s with base-directory set to %s" % (dir,base_dir));
    cmd= "cd %s/%s; lcov --capture --directory . --base-directory %s --output-file %s.out" % (working_dir, dir, base_dir, dir.replace(os.path.sep, '.'))
    (retcode, output)= commands.getstatusoutput(cmd)
    if not retcode == 0:
      logging.error("Failed to capture LCOV data for %s/%s. Exiting." % (working_dir, dir))
      sys.exit(1)
    if dir == 'plugin':
      patch_plugin(working_dir)
    elif dir == 'drizzled':
      patch_drizzled(working_dir)

  os.chdir(working_dir)
  logging.info("Aggregating LCOV info files")
  cmds= ["rm -rf lcov && mkdir -p lcov && lcov"]
  cmds= cmds + [
    "-a %s/%s/%s.out" % (working_dir, dir, dir.replace(os.path.sep, "."))
    for dir in lcov_dirs
  ]
  cmds= cmds + ["-o lcov/all.out"]
  cmd= " ".join(cmds)

  (retcode, output)= commands.getstatusoutput(cmd)

  if not retcode == 0:
    logging.error("Failed to aggregate LCOV files. Got error: %s" % output)
    sys.exit(1)

  strip_include_files(working_dir)
  os.chdir(working_dir)

  # Produce HTML coverage reports
  logging.info("Producing LCOV HTML coverage reports.")

  (retcode, output)= commands.getstatusoutput("genhtml -o lcov lcov/all.out")
  if not retcode == 0:
    logging.error("Failed to generate LCOV HTML coverage report. Got error: %s" % output)
    sys.exit(1)

  if variables['no_rsync'] is False:

    # rsync the reports to drizzle.org
    logging.info("Syncing LCOV results with drizzle.org.")

    (retcode, output)= commands.getstatusoutput("rsync -avz lcov/ %s@drizzle.org:web/lcov/" % variables['ssh']['ssh_user'])

    if not retcode == 0:
      logging.error("Failed to rsync the reports to drizzle.org. Got error: %s" % output)
      sys.exit(1)

  # Scrape the top-level percentages from lcov/index.html and store them in DB
  if variables['no_store_db'] == False:
    logging.info("Storing results of this run in local DB.")

    # Scrape the percentages...
    percentages= {}

    # Grab our index file for the lcov reports
    index_file_contents= open("%s/lcov/index.html" % working_dir).read()
    for dir in lcov_dirs:

      pattern= "href=\"%s/index.html\"" % dir
      pos= index_file_contents.find(pattern)
    
      if pos > 0:
        float_pattern= "alt=\""
        new_pos= index_file_contents.find(float_pattern, pos)
        if new_pos:
          new_pos= new_pos + 5 # length of alt="
          end_pos= new_pos + 1
          while index_file_contents[end_pos].isdigit() or index_file_contents[end_pos] == '.':
            end_pos= end_pos + 1
          dir_percent= float(index_file_contents[new_pos:end_pos])
          percentages[dir]= dir_percent
      else:
        percentages[dir]= 0.0

    # Insert into the DB...
    for dir in lcov_dirs:
      sql= ""
      logging.info("%s %s %s %s %0.2f " % (server_name, server_version, dir, run_date, percentages[dir]))
      sql= sql + """
      INSERT INTO lcov_stats (server, version, dir_name, run_date, coverage_percent) VALUES ('%s','%s','%s','%s', %0.2f); """ % (server_name, server_version, dir, run_date, percentages[dir])

      from drizzle.automation.lib import db
      db.execute_sql(sql)

  return True
