#! /usr/bin/python
# -*- mode: c; c-basic-offset: 2; indent-tabs-mode: nil; -*-
# vim:expandtab:shiftwidth=2:tabstop=2:smarttab:
#
# Copyright (C) 2009 Sun Microsystems
#
# Authors:
#
#  Jay Pipes <joinfu@sun.com>
#  Lee Bieber <lee.bieber@sun.com>
#
# This file is part of the Drizzle Automation Project.
#
# The Drizzle Automation Project is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation.
#
# The Drizzle Automation Project is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with The Drizzle Automation Project (see COPYING.LESSER). If not, 
# see <http://www.gnu.org/licenses/>.

"""Script to automate running random query generator."""

import sys
import os
import os.path
from drizzle.automation.lib  import logging
from drizzle.automation.lib  import util
import commands
import datetime
import time

def execute(processing_mode, variables):
  # Set/verify some required variables, depending on the server
  # we are processing

  working_dir= variables['working_dir']
  run_date= datetime.datetime.now().isoformat()
  server_name= variables['server']
  randgen_home= variables['randgen']['randgen_home']

  os.chdir(working_dir)

  # For BZR mode, the "version" of the server (used in identifying the 
  # command run when logging to the database) is the branch and revision number.
  #
  # For MySQL Sandbox mode, the version is the text after the "msb_" prefix
  # on the sandbox.
  if processing_mode == 'bzr':
    server_version= "%s-%s" % (variables['bzr_branch'], variables['bzr_revision'])
  else:
    server_version= variables['mysql_sandbox'].split('_')[1] # Sandbox names are msb_XXXX (e.g. msb_6011 for MySQL 6.0.11)

  # We must instantiate a server and a client adapter which
  # are used to start/stop the server and to query it. Depending
  # on the type of server we are testing/processing, we may also
  # instantiate a "Builder" object, which is used to build the 
  # server from source...
  if server_name in ['drizzled','drizzle']:

    from drizzle.automation.builder.drizzled import DrizzledBzrBuilder as builder_adapter
    from drizzle.automation.client.drizzledb import DrizzleClient as client_adapter

    # Here, depending on the revision, we use a different adapter
    # for the older drizzleadmin-controlled server...
    if str(variables['bzr_revision']) != 'last:1' and int(variables['bzr_revision']) < 950:
      from drizzle.automation.server.olddrizzled import DrizzledServer as server_adapter
    else:
      from drizzle.automation.server.drizzled import DrizzledServer as server_adapter

  elif server_name in ['mysqld','mysql']:
    from drizzle.automation.builder.mysqld import MySQLBzrBuilder as builder_adapter
    from drizzle.automation.client.mysql import MySQLClient as client_adapter
    from drizzle.automation.server.mysqld import MySQLdServer as server_adapter

  else:
    logging.error("Not yet implemented!")
    sys.exit(1)

  server= server_adapter(variables['working_dir'])
  client= client_adapter(variables['working_dir'], server.getPort())

  if variables['no_build'] is False:
    try:
      configure_options= variables['defaults']['configure_options']
    except KeyError:
      configure_options= ""

    try:
      make_options= variables['defaults']['make_options']
    except KeyError:
      make_options= ""

    builder= builder_adapter(variables['working_dir'], configure_options, make_options) 
    builder.build(variables['force_build'])
  
  # stop any servers that are running
  server.stopAll()

  # clean data directory (var)
  server.clear()

  # get the list of tests that are specified
  # in the configuration file in the randgen_test section
  list_of_tests= variables['randgen_tests']

  server.start()
  # Drop test database if it exists and create a new one
  if not util.drop_and_create_test_database(client):
    server.stop()
    sys.exit(1)

  # change to the randgen repository to run the tests
  os.chdir(randgen_home)

  # loop through the list of tests and run each one
  # start and stop the server each time so we have a clean start each run
  fail_tests= 0
  pass_tests= 0
  for key in list_of_tests.keys():
    # run the test
    cmd= "./gentest.pl --dsn=dbi:drizzle:host=localhost:port=%s:user=root:password="":database=test %s" % (server.getPort(),list_of_tests[key])
    logging.info("Running : %s with %s" % (key, cmd))
    (retcode, output)= commands.getstatusoutput(cmd)
    if not retcode == 0:
      logging.info("%s Failed\n%s" % (key, output))
      fail_tests= fail_tests + 1
      # if a test fails restart the server to get a clean run
      server.stop()
      time.sleep(3)
      server.clear()
      server.start()
      if not util.drop_and_create_test_database(client):
        server.stop()
        sys.exit(1)
    else:
      logging.info("%s Passed" % key)
      pass_tests= pass_tests + 1

  server.stop()

  # output final results, return non-zero status if any test failures
  logging.info("%d tests passed" % pass_tests)
  logging.info("%d tests failed" % fail_tests)
  if fail_tests > 0:
    sys.exit(1)
    
  return True
